# MIT License
#
# Copyright (c) 2021 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from hond.metric import Metric


def test_metric():
    """Metric Tests"""

    metric = Metric(
        "customers.1234567.cpu",
        20.2,
        {"agent": "ebe94451-f111-42f5-9778-da7db3f0bcde"},
        "732a5add-c21e-4423-bded-7a97ff0c8be8"
    )

    assert metric.name == "customers.1234567.cpu"
    assert metric.value == 20.2
    assert metric.meta == {"agent": "ebe94451-f111-42f5-9778-da7db3f0bcde"}
    assert metric.id == "732a5add-c21e-4423-bded-7a97ff0c8be8"
    assert metric.timestamp > 0

    metric = Metric(
        "customers.1234568.memory",
        22.2,
        {},
        None,
        1662503511
    )

    assert metric.name == "customers.1234568.memory"
    assert metric.value == 22.2
    assert metric.meta == {}
    assert metric.id != ""
    assert metric.timestamp == 1662503511



# MIT License
#
# Copyright (c) 2021 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import time

from hond.logger import Logger
from elasticsearch import Elasticsearch


class ElasticSearch:
    """
    ElasticSearch Driver Class
    """

    def __init__(self, connection, index_name):
        """Inits elasticsearch"""
        self.logger = Logger().get_logger(__name__)
        self.client = Elasticsearch(connection)
        self.index_name = index_name

    def get_client(self):
        """
        Get elasticsearch client

        Returns:
            a dict of client info
        """
        return self.client.info()

    def migrate(self, shards=1, replicas=1):
        """
        Create metric index
        """
        doc = {
            "settings": {"number_of_shards": shards, "number_of_replicas": replicas},
            "mappings": {
                "properties": {
                    "id": {"type": "text"},
                    "name": {"type": "text"},
                    "value": {"type": "float"},
                    "timestamp": {"type": "long"},
                    "meta": {"type": "object"},
                }
            },
        }

        response = self.client.index(index=self.index_name, document=doc)

        return response

    def insert(self, metric):
        """
        Insert metrics into elastic search
        """
        doc = {
            "id": metric.id,
            "name": metric.name,
            "value": metric.value,
            "timestamp": metric.timestamp,
            "meta": metric.meta,
        }

        self.logger.debug("Insert metric into elasticsearch: {}", str(metric))

        response = self.client.index(index=self.index_name, document=doc)

        return response

    def is_absent(self, metric_name, for_in_sec=60):
        """
        Check if the metric is absent for x seconds
        """
        query = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response = self.search(query)

        return (
            response["hits"]["total"]["value"] == 0
            and len(response["hits"]["hits"]) == 0
        )

    def equal(self, metric_name, benchmark, for_in_sec=60):
        """
        Check if the metric is equal to the benchmark for x seconds
        """
        query1 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response1 = self.search(query1)

        # If no hits have been found
        if response1["hits"]["total"]["value"] == 0:
            return False

        query2 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {"match_phrase": {"value": {"query": benchmark}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response2 = self.search(query2)

        return (
            response1["hits"]["total"]["value"] == response2["hits"]["total"]["value"]
        )

    def above(self, metric_name, benchmark, for_in_sec=60):
        """
        Check if the metric is above the benchmark for x seconds
        """
        query1 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response1 = self.search(query1)

        # If no hits have been found
        if response1["hits"]["total"]["value"] == 0:
            return False

        query2 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {"range": {"value": {"gt": benchmark}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response2 = self.search(query2)

        return (
            response1["hits"]["total"]["value"] == response2["hits"]["total"]["value"]
        )

    def below(self, metric_name, benchmark, for_in_sec=60):
        """
        Check if the metric is below the benchmark for x seconds
        """
        query1 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response1 = self.search(query1)

        # If no hits have been found
        if response1["hits"]["total"]["value"] == 0:
            return False

        query2 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {"range": {"value": {"lt": benchmark}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response2 = self.search(query2)

        return (
            response1["hits"]["total"]["value"] == response2["hits"]["total"]["value"]
        )

    def above_equal(self, metric_name, benchmark, for_in_sec=60):
        """
        Check if the metric is above or equal the benchmark for x seconds
        """
        query1 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response1 = self.search(query1)

        # If no hits have been found
        if response1["hits"]["total"]["value"] == 0:
            return False

        query2 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {"range": {"value": {"gte": benchmark}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response2 = self.search(query2)

        return (
            response1["hits"]["total"]["value"] == response2["hits"]["total"]["value"]
        )

    def below_equal(self, metric_name, benchmark, for_in_sec=60):
        """
        Check if the metric is below or equal the benchmark for x seconds
        """
        query1 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response1 = self.search(query1)

        # If no hits have been found
        if response1["hits"]["total"]["value"] == 0:
            return False

        query2 = {
            "query": {
                "bool": {
                    "must": [
                        {"match_phrase": {"name": {"query": metric_name}}},
                        {"range": {"value": {"lte": benchmark}}},
                        {
                            "range": {
                                "timestamp": {"gte": int(time.time()) - for_in_sec}
                            }
                        },
                    ]
                }
            }
        }

        response2 = self.search(query2)

        return (
            response1["hits"]["total"]["value"] == response2["hits"]["total"]["value"]
        )

    def search(self, query):
        """
        Query elasticsearch index
        """
        return self.client.search(index=self.index_name, body=query)

    def evaluate(expression):
        """
        Evaluate a trigger value

        Examples:
            m{customers.123.456.789.cpu>=20}[30s]
            m{customers.123.456.789.cpu<20}[30s]
            m{customers.123.456.789.cpu==nul}[30s]
            m{customers.123.456.789.cpu==nul}[30s] and m{customers.123.456.789.mem==nul}[30s]

        TODO: switch to safer way other than eval but right now triggers is not a user input
        """
        result = []
        expressions = re.split(" and | or ", expression)

        for exp in expressions:
            pattern = r"^(m)\{(.*)\}(\[(.*)\])?$"
            match = re.match(pattern, exp)
            if match:
                items = re.split(">=|<=|==|>|<", match.group(2))

                if items[1] == "nul":
                    metric_name = items[0]
                    for_in_sec = int(match.group(4))

                    value = self.is_absent(metric_name, for_in_sec)
                else:
                    metric_name = items[0]
                    benchmark = float(items[1])
                    for_in_sec = int(match.group(4))

                    if "==" in match.group(2):
                        value = self.equal(self, metric_name, benchmark, for_in_sec)
                    elif ">=" in match.group(2):
                        value = self.above_equal(
                            self, metric_name, benchmark, for_in_sec
                        )
                    elif "<=" in match.group(2):
                        value = self.below_equal(
                            self, metric_name, benchmark, for_in_sec
                        )
                    elif ">" in match.group(2):
                        value = self.above(self, metric_name, benchmark, for_in_sec)
                    elif "<" in match.group(2):
                        value = self.below(self, metric_name, benchmark, for_in_sec)

                if value:
                    result.append("True")
                else:
                    result.append("False")
            else:
                raise Exception("Invalid expression: {}".format(exp))

        out = expression

        for i in range(len(expressions)):
            out = out.replace(expressions[i], result[i])

        return eval(out)



# MIT License
#
# Copyright (c) 2021 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import uuid
import json
import time


class Metric:
    """Metric Class"""

    def __init__(self, name, value, meta={}, id=None, timestamp=None):
        self._name = name
        self._meta = meta
        self._value = value
        self._id = str(uuid.uuid4()) if id is None else id
        self._timestamp = int(time.time()) if timestamp is None else timestamp

    @property
    def name(self):
        """
        Gets the metric name

        Returns:
            The metric name
        """
        return self._name

    @property
    def value(self):
        """
        Gets the metric value

        Returns:
            The metric value
        """
        return self._value

    @property
    def id(self):
        """
        Gets the metric id

        Returns:
            The metric id
        """
        return self._id

    @property
    def meta(self):
        """
        Gets the metric meta data

        Returns:
            The metric meta data
        """
        return self._meta

    @property
    def timestamp(self):
        """
        Gets the metric timestamp

        Returns:
            The metric timestamp
        """
        return self._timestamp

    @classmethod
    def from_json(cls, data):
        """
        Get metric from JSON string

        Args:
            data: the JSON string

        Returns:
            An instance of this class
        """
        data = json.loads(data)

        return cls(
            data["name"], data["value"], data["id"], data["meta"], data["timestamp"]
        )

    def __str__(self):
        """
        Convert the Object to string

        Returns:
            A JSON representation of this instance
        """
        return json.dumps(
            {
                "name": self._name,
                "value": self._value,
                "id": self._id,
                "meta": self._meta,
                "timestamp": self._timestamp,
            }
        )



# Copyright 2020 Clivern
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hond.store.elasticsearch import ElasticSearch
from hond.metric import Metric


def main():
    metric = Metric(
        "customers.123.456.789.cpu",
        40.34,
        {"agentId": "1bee4e3c-0976-44d9-bf4a-6432857e4f3c"}
    )
    es = ElasticSearch("http://localhost:9200")
    es.migrate("metric")
    es.insert(metric, "metric")
    es.is_absent("metric", "customers.123.456.789.cpu")


if __name__ == '__main__':
    main()
